#!/bin/sh
#
# create backup of the data / configuration
#
# Usage backup.sh [options] <backup name>
#
#   <backup name> is either a pre-defined or a custom backup configuration name
#
#   options:
#
#      -d        : print debug messages
#      -n        : do not create the backup file, print file and database tables patterns to be used for creating the backup file
#      -t        : simulate creating the backup file, prints the list of the files and database tables to be added to the backup file
#      -s <date> : add changes since <date> only (affects files only). Preferred date format is "%Y-%m-%d %H:%M:%S"
#
#  With no option the backup file will be created
#

. $HOME/etc/setenv.sh

WORK_DIR="$TMP/backup.$$"

#
# Prints usage information and exits
#
# [$1] - error message to be printed
#
function usage_()
{
	[ -n "$1" ] && echo "$1" >&2

	cat <<-ENDOFUSAGE

		Usage:  backup.sh [options] <backup name>

		        <backup name> is either a pre-defined or a custom backup configuration name

		    options:

		        -d        : print debug messages
		        -n        : do not create the backup file, print file and database tables patterns to be used for creating the backup file
		        -t        : simulate creating the backup file, prints the list of the files and database tables to be added to the backup file
		        -s <date> : add changes since <date> only (affects files only). Preferred date format is "%Y-%m-%d %H:%M:%S"

		    With no option the backup file will be created

	ENDOFUSAGE

	[ -n "$1" ] && exit 1

	exit 0
}

#
# cleanup working directory and exit
#
# [$1] - exit code
#
cleanup_()
{
	if [ -d "$WORK_DIR" ] ; then
		rm -r "$WORK_DIR"
		[ -f "${BCK_SQLDUMP_FILE}" ] && rm ${BCK_SQLDUMP_FILE}
		[ -f "${BCK_OSCONFIG_FILE}" ] && rm ${BCK_OSCONFIG_FILE}
		[ -f "${BCK_README_FILE}" ] && rm ${BCK_README_FILE}
	fi  

	exit $1
}


#
# prints the list of database tables to be included in the backup
#
# $1 - content file generated from backup configuration
# $2 - list of all tables
#function getTableList_()
#{
#	local contentFile="$1"
#	local allTables="$2"
#
#
##	"${BCK_CONTENT_FILE}" "${ALL_TABLES}"`
#}

isPrintOnly=""
isSimulate=""
sinceDate=""
isDebug="false"

while getopts 'hdnts:' OPT; do
	case $OPT in
		h) usage_ ;;
		d) isDebug="true" ;;
		n) isPrintOnly="true" ;;
		t) isSimulate="true" ;;
		s) sinceDate="$OPTARG" ;;
		*) usage_ "Invalid option: $OPT" ;;
	esac
done

shift $(( $OPTIND - 1 ))

backupConfig="$1"

[ -z "$backupConfig" ] && usage_ "Missing argument: name of the backup configuration"

[ "true" = "$isPrintOnly" -a "true" = "$isSimulate" ] && usage_ "Only one of the '-n' and '-t' options can be specified"

[ "true" = "$isPrintOnly" ] && echo "Running backup.sh in print only mode for $backupConfig configuration name"
[ "true" = "$isSimulate" ] && echo "Simulate backup.sh for $backupConfig configuration name"

# get site ID for tagging the backup file
siteID=""
[ -f "$CONF/site.conf" ] && . $CONF/site.conf
[ -z "$siteID" ] && siteID="unknownSite"

# get server ID for tagging the backup file
SELF_SERVER_ID=""
[ -f "$CONF/serverid.cfg" ] && . $CONF/serverid.cfg
[ -z "$SELF_SERVER_ID" ] && SELF_SERVER_ID="unknownServer"

dateTime=$(date)
tagDateTime=$(date +"%Y-%m-%d-%H%M%S" -d "$dateTime")

# Replace star in configuration name
tagConfig=`sed -e 's;*;ALL;g' -e 's;\.;_;g' <<<"$backupConfig"`


BCK_TAG="${siteID}.${SELF_SERVER_ID}.${tagDateTime}.${tagConfig}"

BACKUP_TARGZ=backup.${BCK_TAG}.tar.gz

BCK_CONFIG_FILE="$WORK_DIR/config"
BCK_CONTENT_FILE="$WORK_DIR/content"

BCK_FILES_ADD="$WORK_DIR/files.add"
BCK_FILES_EXCLUDE="$WORK_DIR/files.exclude"

BCK_TABLES_ADD="$WORK_DIR/tables.add"
BCK_TABLES_EXCLUDE="$WORK_DIR/tables.exclude"

BCK_SQLDUMP_FILE="$TMP/tables.${BCK_TAG}.dmp"

BCK_OSCONFIG_FILE="$TMP/osconfig.${BCK_TAG}.tar"

BCK_README_RULES="$WORK_DIR/readme.rules"

BCK_README_FILE="$TMP/README.${BCK_TAG}"

ALL_TABLES=`$MYSQL_HOME/bin/mysqlshow $DBNAME | sed -n 's;^| \([^ ]*\) *|$;\1;p'`
if [ $? -ne 0 ] ; then
	echo "ERROR: could not collect database table names" >&2
	exit 1
fi

ALL_FILES="${VAR#$HOME/} ${DATA#$HOME/}"

mkdir $WORK_DIR
if [ $? -ne 0 ] ; then
	echo "ERROR: working directory could not be created" >&2
	exit 1
fi


# echo "Collect files to be backed up"

cat <<-ENDOFALL > ${BCK_CONFIG_FILE}
	ALL FILE ${ALL_FILES} ${BCK_OSCONFIG_FILE#$HOME/}
	ALL DATABASE $(awk -vORS=" " '{print $0}' <<<"${ALL_TABLES}")
	config.os FILE ${BCK_OSCONFIG_FILE#$HOME/}
ENDOFALL

$ETC/rc listbackup 2>/dev/null | sed -n '/^[^#]/p' >> ${BCK_CONFIG_FILE}

# factory default aliases
[ -f "$ETC/backupAliases.cfg" ] && cat $ETC/backupAliases.cfg >> ${BCK_CONFIG_FILE}


# Add custom aliases to backup config file
[ -f "$CONF/backupAliases.cfg" ] && cat $CONF/backupAliases.cfg >> ${BCK_CONFIG_FILE}

CWD=`pwd`

cd $HOME

$ETC/collectBackup.awk -vDebug="$isDebug" -vTarget="$backupConfig" -vFilesAdd="$BCK_FILES_ADD" -vFilesExclude="$BCK_FILES_EXCLUDE" \
		-vTablesAdd="$BCK_TABLES_ADD" -vTablesExclude="$BCK_TABLES_EXCLUDE" -vReadmeRules="$BCK_README_RULES" ${BCK_CONFIG_FILE} >${BCK_CONTENT_FILE}

if [ $? -ne 0 ] ; then
	echo "An error occured while processing backup configuration file(s)" >&2
	cd $CWD
	cleanup_ 1
fi

cd $CWD

# Get tables to be dumped
DUMP_TABLES=""
if [ -f "$BCK_TABLES_ADD" ] ; then
	if [ -f "$BCK_TABLES_EXCLUDE" ] ; then
		DUMP_TABLES=`grep -f $BCK_TABLES_ADD <<<"$ALL_TABLES" | grep -v -f $BCK_TABLES_EXCLUDE`
	else
		DUMP_TABLES=`grep -f $BCK_TABLES_ADD <<<"$ALL_TABLES"`
	fi

	if [ "true" != "${isPrintOnly}" -a "true" != "${isSimulate}" ] ; then
		$MYSQL_HOME/bin/mysqldump --lock-tables --extended-insert --add-drop-table $DBNAME $DUMP_TABLES > $BCK_SQLDUMP_FILE
		echo ${BCK_SQLDUMP_FILE#$HOME/} >> $BCK_FILES_ADD
	else
		echo "Database table(s) to backup:"
		echo "$DUMP_TABLES" | sed 's;^;   ;'
	fi
else
	echo "There are no database tables to backup"
fi


if [ "true" = "${isPrintOnly}" ] ; then
	echo "Files to be included in backup"
	sed 's;^;   ;' $BCK_FILES_ADD
	if [ -f "$BCK_FILES_EXCLUDE" ] ; then
		echo "Files to be excluded from backup"
		sed 's;^;   ;' $BCK_FILES_EXCLUDE
	fi

	cleanup_ 0
fi

tarCmd="tar --create --ignore-failed-read --files-from ${BCK_FILES_ADD}"
[ -f "${BCK_FILES_EXCLUDE}" ] && tarCmd="$tarCmd --exclude-from ${BCK_FILES_EXCLUDE}"
[ -n "${sinceDate}" ] && tarCmd="$tarCmd --newer=\"${sinceDate}\""

if grep -q "^${BCK_OSCONFIG_FILE#$HOME/}$" $BCK_FILES_ADD; then
	# create the os config backup file
	$ASROOT backup.asr "${BCK_OSCONFIG_FILE}" "${BCK_TAG}" "${sinceDate}"
	if [ $? -ne 0 ] ; then
		echo "OS config backup file could not be created" >&2
		cleanup_ 1
	fi
fi


if [ "true" = "${isSimulate}" ] ; then
	cd $HOME
	echo "Content of the backup file would be:"
	eval "$tarCmd --to-stdout 2>/dev/null | tar t | sed 's;^;   ;'"
	[ -n "$DUMP_TABLES" ] && echo "   ${BCK_SQLDUMP_FILE#$HOME/}"

	cd $CWD
	cleanup_ 0
fi

# create the README file:


cat <<-ENDOFREADME > ${BCK_README_FILE}
	Date:               ${dateTime} 
	Site ID:            ${siteID}
	Server ID:          ${SELF_SERVER_ID}
	Server Hostname:    ${SERVERNAME} [$(hostname --long)]
	Server IP address:  ${SERVERIP}

	Backup configuration:  ${backupConfig}
ENDOFREADME

if [ -n "${sinceDate}" ] ; then
	echo "Contains changes since ${sinceDate} only" >> ${BCK_README_FILE}
fi

if [ -s "$BCK_README_RULES" ] ; then
	cat <<-END2OFREADME >> ${BCK_README_FILE}

		========================================================
	
		Rules used for creating the backup file:

	END2OFREADME
	cat $BCK_README_RULES >> ${BCK_README_FILE}
fi

if [ -n "$DUMP_TABLES" ] ; then
	
	cat <<-END3OFREADME >> ${BCK_README_FILE}

		========================================================
	
		Database tables included in the backup:

	END3OFREADME

	echo "$DUMP_TABLES" | sed 's;^;   ;' >> ${BCK_README_FILE}
fi
	

cat <<-END4OFREADME >> ${BCK_README_FILE}

	========================================================

	Files/Directories included to (+) or excluded from (-) the backup:

END4OFREADME

(sed 's;^;+ ;' $BCK_FILES_ADD; sed 's;^;- ;' $BCK_FILES_EXCLUDE) 2>/dev/null | sort -k 2 |  sed 's;^;   ;' >> ${BCK_README_FILE}

echo ${BCK_README_FILE#$HOME/} >> $BCK_FILES_ADD

cd $HOME

eval $tarCmd --gzip --file=$BACKUP_VAR/$BACKUP_TARGZ 2>/dev/null
if [ $? -ne 0 ] ; then
	cd $CWD
	echo "An error occured while creating $BACKUP_TARGZ" >&1
	cleanup_ 1
fi

cd $CWD
cleanup_ 0
