#!/bin/gawk -f
#
# Component: etc
#
# Description:
#    Collect and list files/SQL tables to back up.
#    This script is intended to be used by backup.sh only
#
# Usage:
#
#   Debug           : "true" value indicates to print debug messages
#   Target          : backup alias name or category pattern to be used to create the backup
#   FilesAdd        : filename for file patterns to be added to the backup
#   FilesExclude    : filename for file patterns to be excluded from the backup
#   TablesAdd       : filename for database tables to be added to the backup
#   TableExclude    : filename for database tables to be excluded from the backup
#   ReadmeRules     : backup configuration rules playing role in creating the backup
#   WarnOnly        : do not fail if a required output file parameter is missing, print to standard output
#
# Last Updated     : $Author$
# Update Date Time : $Date$
# RCS File Name    : $RCSfile$
# RCS Revision     : $Revision$
# Locked by        : $Locker$
# Symbolic name    : $Name$
# State            : $State$
#

#
# Print debug messages if debug mode is turned on
#
# str - debug message to be printed
#
function DEBUG(str) {
	if (Debug == "true")
		print "DEBUG: " str;
}

#
# END is always called even if exit called. Set a flag for END to do nothing
#
function QUIT() {
	isQuit = "true";
	exit(1);
}

#
# Checks if a string starts with a minus sign
#
# str - the string to be checked
#
# return true if and only if the first character in str is a minus sign
#
function isMinus(str) {
	return match(str, "^-") != 0;
}


#
# List categories (or category patterns) of an alias
#
# str      - alias name
# isNegate - negate the signing of the categories listed in the alias
#
# return the categories in alias with requested signing
#
function listAliasCategories(str, isNegate) {
	v = "";
	n = split(str, a, " ");
	for (ndx = 1; ndx <= n; ndx++) {
		vname = a[ndx];
		if (vname == "")
			continue;

		isNeg = isMinus(a[ndx]);
		if (isNeg)
			vname = substr(vname, 2);

		isNeg = isNegate ? !isNeg : isNeg;
		if (isNeg) {
			v = v " -" vname;
		}
		else {
			v = v " " vname;
		}
	}

	DEBUG("listAliasCategories(" str ", " isNegate ") returns [" v "]");
	return v;
}


#
# Gets the list of categories matching to a category pattern.
# Pattern may contain '*' to match on any character string.
#
# pattern - category pattern
#
# return the list of categories matching to the category pattern
#
function getMatchingCategories(pattern) {
	regPattern = pattern;
	gsub("^", "^", regPattern);
	gsub("$", "$", regPattern);
	gsub("\\.", "\\.", regPattern);
	gsub("\\*", ".*", regPattern);
	DEBUG("getMatchingCategories(" pattern ") regexp: [" regPattern "]");

	matchingNames = "";
	for (categoryName in allNames) {
		if (allNames[categoryName] != 1)
			continue;
		if (match(categoryName, regPattern)) {
			DEBUG(categoryName " matches to " pattern);
			matchingNames = matchingNames " " categoryName;
		}
	}

	return matchingNames
}


#
# Adds a new rule to the list (array) of rules. If the list already contains the rule, it will not be to the list.
#
# This function is used by expandTarget to list actual rules playing role to
# create the backup file
#
# newRule - rule to be added the the rule list.
#
function addRule(newRule) {
	for (iRules = 1; iRules <= numRules; iRules++) {
		if (tRules[iRules] == newRule) {
			# The rule is already in the list
			return;
		}
	}

	numRules++;
	tRules[numRules] = newRule;
}

#
# Collects and prints the rules (categories and aliases) playing role to create the backup file.
#
# targetRule - backup target
#
function expandTarget(targetRule) {
	numRules = 1;
	ndxRule = 0;

	tRules[1] = targetRule;

	while (++ndxRule <= numRules) {
		curRule = tRules[ndxRule];
	
		curAlias = aliasRule[curRule];
		if (curAlias != "") {
			print curRule " := " curAlias >ReadmeRules;
			print "" >ReadmeRules;
			if (curAlias == curRule) {
				# Built in rule
				continue;
			}

			nAlias = split(curAlias, aAlias, " ");
			for (ndx = 1; ndx <= nAlias; ndx++) {
				vname = aAlias[ndx];
				if (vname == "")
					continue;

				if (isMinus(vname))
					vname = substr(vname, 2);

				addRule(vname);
			}

			continue;
		}

		# Not an alias: category or category pattern
		lstCategories = getMatchingCategories(curRule);
		nCategory = split(lstCategories, aCategory, " ");
		if (nCategory == 1 && aCategory[1] == curRule) {
			# Rule is an exact category, not a pattern. Do not print it
			continue;
		}

		print curRule " :" >ReadmeRules;
		for (ndx = 1; ndx <= nCategory; ndx++) {
			vname = aCategory[ndx];
			if (vname == "")
				continue;

			print "  " vname >ReadmeRules;
		}
		print "" >ReadmeRules;
	}

	targetList = alias[targetRule];
	if (targetList != "") {
		print "" >ReadmeRules
		print "The expanded rule is:" >ReadmeRules
		print "" >ReadmeRules
		print "  " targetRule " := " targetList >ReadmeRules;
		print "" >ReadmeRules
	}
}

#######################################################


BEGIN {
	# for collecting defined values
	file["DEFINED"] = "";
	database["DEFINED"] = "";

	# Built in aliases
	alias["ALL"] = "ALL";
	alias["DEFINED"] = "DEFINED";

	# Alias rules
	aliasRule["ALL"] = "";
	aliasRule["DEFINED"] = "";

	
	allNames["ALL"] = 1;
	allNames["DEFINED"] = 1;



	
	if (FilesAdd == "") {
		print "FilesAdd (output file for file patterns to be added to the backup) is not specified." >"/dev/stderr";
		if (WarnOnly != "true")
			QUIT();
		FilesAdd = "/dev/stdout";
	}
	if (FilesExclude == "") {
		print "FilesExclude (output file for file patterns to be excluded from the backup) is not specified." >"/dev/stderr";
		if (WarnOnly != "true")
			QUIT();
		FilesExclude = "/dev/stdout";
	}
	if (TablesAdd == "") {
		print "TablesAdd (output file for database tables to be added to the backup) is not specified." >"/dev/stderr";
		if (WarnOnly != "true")
			QUIT();
		TablesAdd = "/dev/stdout";
	}
	if (TablesExclude == "") {
		print "TablesExclude (output file for database tables to be excluded from the backup) is not specified." >"/dev/stderr";
		if (WarnOnly != "true")
			QUIT();
		TablesExclude = "/dev/stdout";
	}
	if (ReadmeRules == "") {
		print "ReadmeRules (output file for rules used to create the backup file) is not specified." >"/dev/stderr";
		if (WarnOnly != "true")
			QUIT();
		ReadmeRules = "/dev/stdout";
	}
}


# Skip comment lines
/^[:space:]*#/ {
	DEBUG("skip comment line: $0");
	next;
}

# Process file definitions
($2 == "FILE") {
	DEBUG("FILE " $1);
	curValue = file[$1];
	for (i = 3; i <= NF; i++) {
		curValue = curValue " " $i;
		if ($1 != "ALL")
			file["DEFINED"] = file["DEFINED"] " " $i;
	}

	file[$1] = curValue;
	allNames[$1] = 1;
	next;
}

# Process database table definitions
($2 == "DATABASE") {
	DEBUG("DATABASE " $1);
	curValue = database[$1];
	for (i = 3; i <= NF; i++) {
		curValue = curValue " " $i;
		if ($1 != "ALL")
			database["DEFINED"] = database["DEFINED"] " " $i
	}

	database[$1] = curValue;
	allNames[$1] = 1;
	next;
}

# Process alias definitions
($2 == "ALIAS") {
	DEBUG("ALIAS " $1);
	if (match($1, "\\.") != 0) {
		print "Invalid ALIAS name: $1; Alias names may not contain dots (.)" >"/dev/stderr" 
		QUIT();
	}

	if (alias[$1] != "") {
		print "ALIAS $1 is already defined" >"/dev/stderr" 
		QUIT();
	}

	rules = "";
	referencedNames = "";
	for (i = 3; i <= NF; i++) {
		DEBUG("   #" i ": " $i);
		name = $i;
		rules = rules " " name;
		isExclude = isMinus($i);
		if (isExclude) {
			name = substr($i, 2);
			DEBUG("is exclude " name "[" $i "]");
		}

		if (allNames[name] == 1) {
			# is a defined name

			if (alias[name] == "") {
				# a category name; leave it as is
				referencedNames = referencedNames " " $i;
			}
			else {
				# an alias variable; expand variable refernces in alias
				referencedNames = referencedNames listAliasCategories(alias[name], isExclude);
			}
				
			continue;
		}
		
		# category pattern

		varList = getMatchingCategories(name);
		if (varList == "") {
			# No such configuration(s) - return an error
			print "Reference to an undefined category or alias in ALIAS (" $1 ") definition: " name >"/dev/stderr";
			QUIT();
		}

		referencedNames = referencedNames listAliasCategories(varList, isExclude);
	}
	
	aliasRule[$1] = rules;
	alias[$1] = referencedNames;

	DEBUG(" alias[" $1 "] := " referencedNames);
	DEBUG("ADD " $1 " to all names");
	allNames[$1] = 1;
}



END {
	if (isQuit == "true")
		exit(1);
	
	targetList = alias[Target];
	
	if (targetList == "") {
		# target is not an alias name; might be a category pattern
		targetList = getMatchingCategories(Target);
	}

	DEBUG("TARGET: " Target " [" targetList "]");
	
	if (targetList == "") {
		print "Unrecognized backup configuration(category or alias) name: " Target >"/dev/stderr";
		exit(1);
	}

	
	nTargets = split(targetList, aTarget, " ");
	for (ndxTarget = 1; ndxTarget <= nTargets; ndxTarget++) {
		targetItem = aTarget[ndxTarget];

		if (targetItem == "")
			continue;
	
		isExclude = isMinus(targetItem);
		if (isExclude) {
			targetItem = substr(targetItem, 2);
			DEBUG("is exclude " targetItem " [" targetItem "]");
		}

		files = file[targetItem];
		tables = database[targetItem];

		DEBUG(targetItem "(" isExclude ") [" files "] [" tables "]");

		if (files != "") {
			nFiles = split(files, aFiles, " ");
			for (iFiles = 1; iFiles <= nFiles; iFiles++) {
				fileName = aFiles[iFiles];
				if (fileName == "")
					continue;
				resultFILES[fileName] = !isExclude;
			}
		}

		if (tables != "") {
			nTables = split(tables, aTables, " ");
			for (iTables = 1; iTables <= nTables; iTables++) {
				tableName = aTables[iTables];
				if (tableName == "")
					continue;
				resultTABLES[tableName] = !isExclude;
			}
		}
	}

	for (f in resultFILES) {
		if (resultFILES[f] == 1) {
			DEBUG("FILE + " f);
			wasFile = 0;
			cmd_ls = "ls -d " f " 2>/dev/null";
			while ((cmd_ls | getline fline)) {
				print fline > FilesAdd
				wasFile = 1;
			}
			close(cmd_ls);

			if ( !wasFile ) {
				# Files do not exist at collectBackup time (eg. OS level backup file generated on demand)
				print f > FilesAdd
			}
		}
		else {
			DEBUG("FILE - " f);
			print f > FilesExclude
		}
	}

	for (t in resultTABLES) {
		if (resultTABLES[t] == 1) {
			DEBUG("TABLE + " t);
			print t > TablesAdd
		}
		else {
			DEBUG("TABLE - " t);
			print t > TablesExclude
		}
	}

	expandTarget(Target);
}
